
(An example of output generated by `example-generate-fun-with-mastermind.dart`.)

# Fun with Mastermind

In this demonstration we investigate the average number of guesses it
takes to win a game of Mastermind using a simple algorithm:

1. We choose a random code from among the potential codes, and guess it.

2. We use the clues we receive to determine whether we have won, in which case we stop and record the number of moves it took, or not, in which case we use the clues to reduce the number of potential codes and go back to step 1.

We are investigating the version of the game that has six colors (which we arbitrarily refer to as A, B, C, D, E and F) and four positions.

Clues are represented by the characters `*` (for black, which means that a color is in the correct position) and `o` (for white, which means a color not already accounted for is in an incorrect position).

We will let the computer play 10000 games; we will only display the first and last 5 games.
  
Here goes:

## The games...
  
  

### Game 1

The secret code: CFBF

|Guess number|Guess|Clues|
|:--:|:--|:--|
|1|FBCB|`  ooo `|
|2|BAFC|`  ooo `|
|3|CFBF|` ****  `|

Solved in 3 guesses.

### Game 2

The secret code: BCDA

|Guess number|Guess|Clues|
|:--:|:--|:--|
|1|DFCB|`  ooo `|
|2|BDFE|` * o `|
|3|CBFA|` * oo `|
|4|ACFD|` * oo `|
|5|BCDA|` ****  `|

Solved in 5 guesses.

### Game 3

The secret code: EDCC

|Guess number|Guess|Clues|
|:--:|:--|:--|
|1|ACCE|` * oo `|
|2|AEBC|` * o `|
|3|ECDC|` ** oo `|
|4|EDCC|` ****  `|

Solved in 4 guesses.

### Game 4

The secret code: BADF

|Guess number|Guess|Clues|
|:--:|:--|:--|
|1|DDCD|`  o `|
|2|ACEA|`  o `|
|3|FBFC|`  oo `|
|4|EFDF|` **  `|
|5|BADF|` ****  `|

Solved in 5 guesses.

### Game 5

The secret code: ECCF

|Guess number|Guess|Clues|
|:--:|:--|:--|
|1|ADAD|`   `|
|2|CECE|` * oo `|
|3|ECCB|` ***  `|
|4|ECCC|` ***  `|
|5|ECCF|` ****  `|

Solved in 5 guesses.

### Game 9996

The secret code: CABC

|Guess number|Guess|Clues|
|:--:|:--|:--|
|1|DACF|` * o `|
|2|EAED|` *  `|
|3|CAAC|` ***  `|
|4|AAAC|` **  `|
|5|CABC|` ****  `|

Solved in 5 guesses.

### Game 9997

The secret code: CACD

|Guess number|Guess|Clues|
|:--:|:--|:--|
|1|FFEF|`   `|
|2|DCAD|` * oo `|
|3|BDAC|`  ooo `|
|4|CCDA|` * ooo `|
|5|CACD|` ****  `|

Solved in 5 guesses.

### Game 9998

The secret code: BBDB

|Guess number|Guess|Clues|
|:--:|:--|:--|
|1|DBCA|` * o `|
|2|EDEA|`  o `|
|3|BFCE|` *  `|
|4|ACCC|`   `|
|5|BBDD|` ***  `|
|6|BBDB|` ****  `|

Solved in 6 guesses.

### Game 9999

The secret code: DFCD

|Guess number|Guess|Clues|
|:--:|:--|:--|
|1|AEAE|`   `|
|2|FBDF|`  oo `|
|3|DCBB|` * o `|
|4|CFBC|` * o `|
|5|DFCD|` ****  `|

Solved in 5 guesses.

### Game 10000

The secret code: AEEA

|Guess number|Guess|Clues|
|:--:|:--|:--|
|1|ADBC|` *  `|
|2|EEBF|` * o `|
|3|FFBB|`   `|
|4|CEEC|` **  `|
|5|AEEA|` ****  `|

Solved in 5 guesses.
## Summary

In 10000 games, the computer found the code in at least 1 guesses, at most 8 guesses, and on average 4.643 guesses.
## trotter

This was just a little demonstration of a real-world application using the trotter library.

Notice that Mastermind codes are permutations with replacement (referred to as `Amalgams` in the trotter library). In this program a single line of code generates a list of all possible Mastermind codes:

```dart
var potentials = Amalgams(4, characters("ABCDEF"))().toList();
```

The game *Bulls & Cows* is played similarly to Mastermind but with non zero digits instead of colors and usually without replacement. Notice that this program would only need a few alterations to investigate that game, including something along the lines of:

```dart
var potentials = Permutations(4, characters("123456789"))().toList();
```

Thanks for your interest! Please let me know if you find any interesting use cases for this library!

Richard Ambler rambler@ibwya.net

  
